module Tabela where

import Data.List
import Data.Semigroup
import Data.Foldable

class Contravariant f where
  -- Functor contravariante, quando a função
  -- necessária aponta para o lado errado.
  contramap :: (b -> a) -> f a -> f b

-- Uma tabela é uma lista de colunas 
newtype Tabela a = Tabela [Coluna a]

-- Uma coluna contém um cabeçalho e uma função que extrai
-- o conteúdo de um registro.
data Coluna a = Coluna { _cabeçalho :: String
                       , _conteudo  :: a -> String
                       }

instance Semigroup (Tabela a) where
  (Tabela t1) <> (Tabela t2) = Tabela (t1<>t2)

instance Monoid (Tabela a) where
  mempty = Tabela []


-- Tenho: _conteudo :: a -> String
-- Quero: _conteudo :: b -> String 
-- Se você me der um b -> a, eu converto o b para a e aplico
-- meu conteudo atual.
instance Contravariant Coluna where
  contramap f col = col { _conteudo = _conteudo col . f }

-- apenas aplica o contramap na tabela 
instance Contravariant Tabela where
  contramap f (Tabela t) = Tabela $ fmap (contramap f) t

-- cria uma tabela
tabela :: String -> (a -> String) -> Tabela a
tabela nome conteudo = Tabela [Coluna nome conteudo]

-- converte a tabela para o formato markdown:
--
--  | cabeçalho1 | cabeçalho2 |
--  |------------|------------|
--  |  valor1_1  | valor1_2   |
--  |  valor2_1  | valor2_2   |
--
tabular :: Tabela a -> [a] -> String
tabular (Tabela colunas) pessoas = unlines
                                 $ map (envolve.padded)
                                 $ cabeçalho:separadores:valores
  where
    cabeçalho = map _cabeçalho colunas
    valores   = map geraLinhas pessoas
    largura   = maxLarguraCols (cabeçalho : valores)

    padded       = zipWith pad largura
    separadores  = map (`replicate` '-') largura
    geraLinhas p = map (`_conteudo` p) colunas

-- * Funções auxiliares

-- envolve a string entre separadores
envolve :: [String] -> String
envolve strings = concat ["|", intercalate "|" strings, "|"]

pad :: Int -> String -> String
pad n word
  | diff <= 0 = word
  | otherwise = word ++ replicate diff ' '
  where
    diff = n - length word

maxLarguraCols :: Foldable t => [[t a]] -> [Int]
maxLarguraCols xss = map (getMax.fold)
                   $ transpose
                   $ map (map (Max . length)) xss
